'' =================================================================================================
''
''   File....... jm_hc-06_configurator__17-03-21.spin
''   Purpose.... Configuration utility for HC-06 module
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2015-2017 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 21 MAR 2017
''
'' =================================================================================================

{{

  Source: http://wiki.pinguino.cc/index.php/SPP_Bluetooth_Modules

  HC-06 AT commands
  
   * AT commands can only be sent to the device using a wired, not wireless, connection.
  
   * The bluetooth device will only respond to AT commands before it is paired/connected.
  
   * AT commands do not need a terminating CR/LF sequence.
  
   * AT commands must be entered quickly (cut and paste or use macro).

   
  The HC-06 firmware has a very limited subset of AT commands as listed below:
  
   * AT responds OK

   * AT+BAUDn where n = 1..C
     1=1200, 2=2400, 3=4800, 4=9600 (default), 5=19200, 6=38400, 7=57600, 8=115200, 9=230400,
     A=460800, B=921600, C=1382400 
     -- Responds OKnnnn. Retained across power offs.
     
   * AT+NAMEname where name 20 or fewer characters. Responds OKname. Retained across power offs.
     -- enter name inside double quotes if spaces are included, e.g., "JM HC-06 #3"
   
   * AT+PINnnnn sets the pairing password. Responds OKsetpin.
   
   * AT+Pn where n is N (no parity), O (odd parity) or E (even parity). Firmware version higher than Linvor 1.5 only.
   
   * AT+VERSION responds with the firmware version.


   Ensure that Bluetooth module is not connected to any other device -- will not enter
   programming mode when connected.   

}}


con { timing }

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' use 5MHz crystal

  CLK_FREQ = _xinfreq * (_clkmode >> 6)                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us

  TRM_BAUD = 115_200                                            ' terminal baud setting
  BT_BAUD  =   9_600                                            ' starting bluetooth baud setting
                                                         
                                                                 
con  { io pins }                                  
                                                                  
  RX1 = 31                                                      ' programming / terminal
  TX1 = 30                                                    
                                                                  
  SDA = 29                                                      ' eeprom / i2c
  SCL = 28                                                    
  
  ' bluetooth connection                                              

  RX2 =  5                                                      ' HC-8+ OPTION.A0
  TX2 =  4                                                      ' HC-8+ OPTION.SM


obj

' main                                                          ' * master Spin cog
  time   : "jm_time_80"                                         '   timing and delays
  parser : "jm_parser"                                          '   parsing engine
  buffer : "jm_textbuf"                                         '   text buffer  
  io     : "jm_io_basic"                                        '   essential io
  term   : "jm_fullduplexserial"                                ' * serial for terminal
  btio   : "jm_fullduplexserial"                                ' * serial for bluetooth
                                                         
' * uses cog when loaded                                           
                                                                  
                                                                  
var

  byte  textbuf[80]


dat { valid characters and tokens }

  CHAR_SET      byte    "-_0123456789"                          ' valid characters for this token set
                byte    "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                byte    "abcdefghijklmnopqrstuvwxyz", 0   
  
  TOKEN_LIST    byte    "SCAN", 0                               ' scan for bt module
                byte    "AT", 0                                 ' valid tokens
                byte    "BAUD", 0    
                byte    "NAME", 0    
                byte    "PIN", 0
                byte    "VERSION", 0
                byte    "SLAVE", 0
                byte    "MASTER", 0
                byte    "HELP", 0


con { enumerated token indexes }

  #0, T_SCAN, T_AT, T_BAUD, T_NAME, T_PIN, T_VERSION, T_SLAVE, T_MASTER, T_HELP

  TOKEN_COUNT = T_HELP+1                                         ' total # of known tokens

                                                                          
pub main | c                                                    
                                                                  
  setup

  time.pause(2000)
  
  term.tx(term#CLS)
  term.str(string("HC-06 Bluetooth Configurator", 13, 13))

  term.rxflush                                                  ' clear any start-up trash
  btio.rxflush                                                   

  scan_bluetooth  
                                                              
  repeat                                                         
    c := term.rx                                                ' get char from terminal
    if (parser.enqueue(c))                                      ' send to parser, if tokens ready
      parser.ucase(0)                                           ' convert command token to upper case
      process_cmd                                               ' process valid tokens
      parser.reset                                              ' reset parser for next command


dat { acceptable bluetooth baud rates }

  BT_RATES      long    1200, 2400, 4800, 9600, 19200, 38400, 57600
                long    115200, 230400, 460800, 921600, 1382400                   

      
pub scan_bluetooth | idx, baud, check

  term.str(string("Scanning..."))

  repeat idx from 0 to 11
    baud := BT_RATES[idx]
    term.str(string(13, "-- "))
    term.dec(baud)

    btio.start(RX2, TX2, %0000, baud)                           ' reset bt uart speed
    time.pause(10)                                              ' allow it to get ready
    btio.rxflush                                                ' flush rx buffer

    btio.tx("A")                                                ' ping Bluetooth at baud
    btio.tx("T")

    check := btio.rxtime(1000)                                  ' allow 1s to respond
    if (check == "O")
      term.str(string("  Found!", 13))  
      btio.rxflush
      return

  ' report error and stop program

  term.tx(term#CLS)
  term.str(string("Error: Could not connect to Bluetooth module"))

  repeat
    waitcnt(0)
                                                                

pub process_cmd | tidx

  tidx := parser.get_token_id(parser.token_addr(0))             ' get index of command token
  
  case tidx
    T_SCAN    : scan_bluetooth                                  ' handle valid tokens
    T_AT      : ping
    T_BAUD    : set_baud
    T_NAME    : set_name
    T_PIN     : set_pin
    T_VERSION : show_version
    T_SLAVE   : set_slave
    T_MASTER  : set_master                                      ' requires HC-06 version 1.7 firmware
    T_HELP    : show_help
    other     : invalid_cmd


pub ping

'' Verifies HC-06 connected at correct baud rate
'' -- HC-06 responds with "OK"

  if (parser.token_count <> 1)                                  ' check token count
    term.str(@UseHdr)
    term.str(@ATMsg)
    return

  set_and_get(string("AT"))

        
pub set_baud | check, baud, idx

'' Sets new baud rate into HC-06
'' -- limited to FDS-friendly baud rates
'' -- restarts serial connection to bt module at new baud rate

  if (parser.token_count <> 2)                                  ' check token count
    term.str(@UseHdr)
    term.str(@BaudMsg)
    return

  check := parser.token_is_number(1)                            ' must enter number

  if (check)
    baud := parser.token_value(1)                               ' extract target baud
    check := false                                              ' assume failure
    repeat idx from 1 to 12                                     ' loop through baud rates
      if (baud == BT_RATES[idx-1])                              ' check
        check := idx                                            ' if found, set index
        quit

  if (check == false)
    term.str(string("Invalid baud setting", 13))
    return

  ' set new baud in module and connection

  buffer.clear                                                  ' clear buffer
  buffer.puts(string("AT+BAUD"))                                ' set command string
  buffer.putc(CHAR_SET[check + 2])                              ' use hex char
                                                                 
  set_and_get(buffer.address)                                   ' update module

  term.str(string("-- resetting BT baud", term#CR))
  
  btio.start(RX2, TX2, %0000, baud)                             ' reset serial link
  time.pause(50)

  set_and_get(string("AT"))      


pub set_name | n, tidx

'' Sets name in HC-06

  if (parser.token_count < 2)                                   ' check token count
    term.str(@UseHdr)                                            
    term.str(@NameMsg)                                           
    return                                                       
                                                                 
  buffer.clear                                                  ' clear buffer
  buffer.puts(string("AT+NAME"))                                ' add command
  buffer.puts(parser.token_addr(1))                             ' add name
                                                                 
  ' name may have spaces which creates multiple tokens           
                                                                 
  n := parser.token_count - 2                                   ' n is # of additional tokens
  tidx := 2                                                     ' first additional is #2
  repeat while (n > 0)                                          ' if any
    buffer.putc(" ")                                            ' add space pad
    buffer.puts(parser.token_addr(tidx++))                      ' add token
    --n                                                          
                                                                 
  buffer.truncate(7 + 20)                                       ' limit name to 20 chars
                                                                 
  set_and_get(buffer.address)                                   ' update module
                                                                 
        
pub set_pin | p_pin, len

'' Set PIN (pairing passcode) in HC-06

  if (parser.token_count <> 2)                                  ' check token count
    term.str(@UseHdr)
    term.str(@PinMsg)
    return

  buffer.clear
  buffer.puts(string("AT+PIN"))
  buffer.puts(parser.token_addr(1))

  set_and_get(buffer.address)                                   ' update module


pub show_version

'' Ask HC-06 firmware version
'' -- response something like "OKlinvorV1.8"

  if (parser.token_count <> 1)                                  ' check token count
    term.str(@UseHdr)
    term.str(@VersionMsg)
    return

  set_and_get(string("AT+VERSION"))


pub set_slave

'' Sets HC-06 to slave mode
'' -- default configuration for HC-06

  if (parser.token_count <> 1)                                  ' check token count
    term.str(@UseHdr)
    term.str(@SlaveMsg)
    return

  set_and_get(string("AT+ROLE=S"))


pub set_master | k                                              ' NOT RECOMMENDED -- NO CONTROL OVER PAIRING

'' Sets HC-06 to master mode

  if (parser.token_count <> 1)                                  ' check token count
    term.str(@UseHdr)
    term.str(@MasterMsg)
    return

  term.tx(term#CR)
  term.str(string("Are you sure you want to set MASTER mode? (Y/N) "))
  term.rxflush
  k := term.rx
  ifnot ((k == "Y") or (k == "y"))
    return

  set_and_get(string("AT+ROLE=M"))

    
dat { help strings }

  UseHdr        byte    "Usage: ", 0

  ScanMsg       byte    "SCAN <CR>", 13, 0            
  ATMsg         byte    "AT <CR>", 13, 0
  BaudMsg       byte    "BAUD newbaud <CR>", 13, 0
  NameMsg       byte    "NAME newname <CR>", 13, 0
  PinMsg        byte    "PIN newcode <CR>", 13, 0
  VersionMsg    byte    "VERSION <CR>", 13, 0
  SlaveMsg      byte    "SLAVE <CR>", 13, 0  
  MasterMsg     byte    "MASTER <CR>", 13, 0                    ' not recommended! 

    
pub show_help

'' List available commands

  term.tx(term#CR)
  term.str(@ScanMsg) 
  term.str(@ATMsg)
  term.str(@BaudMsg)
  term.str(@NameMsg)
  term.str(@PinMsg)
  term.str(@VersionMsg)   
  
  
pub invalid_cmd

  term.str(string(term#CR, "Invalid command: "))
  term.str(parser.queue_addr)                                   ' show me what was bogus
  term.tx(term#CR)


pub set_and_get(p_str)

'' Send string at p_str to HC-06 and wait for response

  term.tx(term#CR)
  term.str(p_str)
  term.tx(term#CR)
  
  btio.str(p_str) 
  get_bt


pub get_bt | c

'' RX and display chars from HC-06 module

  repeat
    c := btio.rxtime(1000)                                      ' wait up to 1s
    if (c => 0)                                                 ' good char
      term.tx(c)                                                ' print to terminal
    else
      quit                                                      ' no char, timeout
    
  term.tx(term#CR)

                             
pub setup

  time.start                                                    ' setup timing & delays
                                                                 
  io.start(0, 0)                                                ' clear all pins (master cog)                                                       

  parser.start(@CHAR_SET, @TOKEN_LIST, TOKEN_COUNT, true)       ' setup parsing engine

  buffer.use(@textbuf, 79)                                      ' setup text buffer
                                                            
  term.start(RX1, TX1, %0000, TRM_BAUD)                         ' start serial for terminal
                                                                  
  time.pause(2000)                                              ' let everything get ready
                                                                  
                              
dat { license }                                                   
                                                                  
{{

  Copyright (c) 2015-2016 Jon McPhalen                                                               

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}