'' =================================================================================================
''
''   File....... jm_textbuf.spin
''   Purpose.... Text buffer object for manipulating strings
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (C) 2015-2017 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started....
''   Updated.... 21 MAR 2017
''
'' =================================================================================================


var

  long  p_buffer                                                ' address of buffer
  long  bsmax                                                   ' size of buffer
  byte  bidx                                                    ' buffer index (next char)
                                                                 
                                                                 
pub null                                                         
                                                                 
  ' This is not a top-level object                               


pub use(p_buf, maxlen)

'' Assign external buffer to object
'' -- p_buf is hub address of buffer
'' -- maxlen is maximum buffer length (without nul)
''    * external buffer must be declared as size maxlen+1
'' -- does not reset buffer; allows on-the-fly switching

  longmove(@p_buffer, @p_buf, 2)                                ' copy pointer and max length
  bidx := strsize(p_buffer)                                     ' update bidx to end

                                                             
pub clear                                                        
                                                                 
'' Clears buffer and resets buffer index                         

  bytefill(p_buffer, 0, bsmax)                                  ' fill buffer with zeros
  bidx := 0


pub set(p_str) | n

'' Sets assigned buffer
'' -- replaces anything already in buffer
'' -- truncates input if required

  n := strsize(p_str) <# bsmax                                  ' number of bytes in source string

  bytemove(p_buffer, p_str, n)                                  ' move source string into buffer
  byte[p_buffer][n] := 0                                        ' terminate
  bidx := n                                                     ' update bidx


pub copy_to(p_str) | n

'' Copies the contents of buffer to array at p_str
'' -- output string must be sized hold contents of buffer

  bytemove(p_str, p_buffer, bidx+1)                             ' copy contents + 0
  

pub putc(c)

'' Add character to end of buffer

  if (bidx < bsmax)                                             ' if space in buffer
    byte[p_buffer][bidx++] := c                                 '  add new character
    byte[p_buffer][bidx] := 0                                   '  re-terminate
                                                                 
                                                                 
pub puts(p_str) | n                                                 
                                                                 
'' Add string to end of buffer
'' -- input string is terminated if too long for buffer

  n := strsize(p_str) <# (bsmax - bidx)                         ' get max chars to move

  bytemove(p_buffer+bidx, p_str, n)                             ' add string to buffer
  bidx += n                                                     ' update bidx
  byte[p_buffer][bidx] := 0                                     ' fix termination                                  

  
pub dec(value) : zflag | divisor, nx                            ' borrowed from FDS

'' Add decimal number to buffer

  nx := (value == negx)                                 
  if (value < 0)                                        
    value := ||(value+nx)                               
    putc("-")                                           
                                                        
  divisor := 1_000_000_000                              
                                                        
  repeat 10                                             
    if (value => divisor)                               
      putc(value / divisor + "0" + nx*(divisor == 1))    
      value //= divisor                                 
      zflag := true                                     
    elseif (zflag or (divisor == 1))                    
      putc("0")                                         
    divisor /= 10


pub rjdec(val, width, pchar) | tmpval, pad

'' Add right-justified decimal value to buffer
'' -- val is value to print
'' -- width is width of (padded) field for value
'' -- pchar is [leading] pad character (usually "0" or " ")

'  Original code by Dave Hein
'  -- modifications by Jon McPhalen

  if (val => 0)                                                 ' if positive
    tmpval := val                                               '  copy value
    pad := width - 1                                            '  make room for 1 digit
  else                                                           
    if (val == negx)                                            '  if max negative
      tmpval := posx                                            '    use max positive for width
    else                                                        '  else
      tmpval := -val                                            '    make positive
    pad := width - 2                                            '  make room for sign and 1 digit
                                                                 
  repeat while (tmpval => 10)                                   ' adjust pad for value width > 1
    pad--                                                        
    tmpval /= 10                                                 
                                                                 
  repeat pad                                                    ' print pad
    putc(pchar)                                                    
                                                                 
  dec(val)                                                      ' print value


pub bin(value) : zflag | idx, c

'' Add variable-width binary value to buffer

  repeat idx from 31 to 0
    c := (value <-= 1) & 1 + "0"
    if (c <> "0")
      zflag := true
    if ((zflag) or (idx == 0))
      putc(c)


pub xbin(value, digits)                                         ' borrowed from FDS 

'' Add fixed-width binary number to buffer

  value <<= 32 - digits
  repeat digits
    putc((value <-= 1) & 1 + "0")


pub hex(value) : zflag | idx, c

'' Add variable-width hex value to buffer

  repeat idx from 7 to 0
    c := lookupz((value <-= 4) & $F : "0".."9", "A".."F")
    if (c <> "0")
      zflag := true
    if ((zflag) or (idx == 0))
      putc(c)
      

pub xhex(value, digits)                                         ' borrowed from FDS 

'' Add fixed-width hex number to buffer

  value <<= (8 - digits) << 2
  repeat digits
    putc(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))


pub equal(p_str)

'' Return true if buffer is equal to string at p_str
'' -- case sensitive

  return strcomp(p_buffer, p_str)


pub nc_equal(p_str)

  return nc_strcomp(p_str)


pub nc_strcomp(p_str) | p_buf

'' Return true if buffer is equal to string at p_str
'' -- not case sensitive

  if (size <> strsize(p_str))                                   ' must be seme length
    return false

  p_buf := p_buffer

  repeat bidx
    if (upper(byte[p_str++]) <> upper(byte[p_buf++]))
      return false 
                                                      
  return true


pub str_index(p_list, lcount) | idx

'' Returns index of string (at p_str) in list of strings (at p_list)
'' -- lcount is number of strings in list
'' -- search is case sensitive

  idx := 0

  repeat lcount
    if (strcomp(p_buffer, p_list))
      return idx
    else
      ++idx
      p_list += strsize(p_list) + 1

  return -1
  

pub upper(c)

'' Convert c to upper case
'' -- does not affect buffer

  if ((c => "a") and (c =< "z"))
    c -= 32

  return c
  

pub ucase | idx

'' Converts alpha characters in buffer to upper case
                    
  repeat idx from 0 to bidx
    byte[p_buffer][idx] := upper(byte[p_buffer][idx])


pub lower(c)

'' Convert c to lower case
'' -- does not affect buffer

  if ((c => "A") and (c =< "Z"))
    c += 32

  return c

    
pub lcase | idx

'' Converts alpha characters in buffer to lower case
                    
  repeat idx from 0 to bidx
    byte[p_buffer][idx] := lower(byte[p_buffer][idx]) 


pub truncate(n)

'' Sets index of buffer to n
'' -- only if buffer longer than n

  if ((n => 0) and (n < bidx))
    bidx := n    
    bytefill(byte[p_buffer][n], 0, bsmax+1-n)
    return true

  return false


pub backspace

'' Deletes last character in buffer
'' -- returns length of buffer

  if (bidx > 0)
    byte[p_buffer][bidx] := 0
    bidx -= 1

  return bidx


pub address

'' Provides buffer address for external string methods

  return p_buffer


pub size

'' Returns size of buffer

  bidx := strsize(p_buffer)                                     ' update

  return bidx


con

  ' Simple crypto
  ' -- for fun, not security!


pub mr_robot(p_str) | len, idx, c                               ' eps2.9_pyth0n-pt1.p7z

'' Encrypt/decrypt buffer using ROT-13 cypher
'' -- see http://www.decode.org/

  caesar(p_str, 13)
  

pub caesar(p_str, key) | idx, c

'' Encrypt/decrypt buffer using Caeser cipher
'' -- p_str is pointer to string to encode
''    * use -1 for buffer assigned to object
'' -- key is 1 to 26
''    * decryption key is 26-key

  if (p_str < 0)
    p_str := p_buffer                                           ' use assigned buffer

  idx := 0
    
  repeat strsize(p_str)
    c := byte[p_str][idx]
    if ((c => "A") and (c =< "Z"))
      c := "A" + ((c-"A") + key) // 26  
    elseif ((c => "a") and (c =< "z"))
      c := "a" + ((c-"a") + key) // 26  
    byte[p_str][idx++] := c    


pub rot47(p_str) | idx, c

'' Encrypt/decrypt buffer using ROT-47 cipher
'' -- p_str is pointer to string to encode
''    * use -1 for buffer assigned to object
'' -- includes punctuation and other characters

  if (p_str < 0)
    p_str := p_buffer                                           ' use assigned buffer

  idx := 0

  repeat strsize(p_str)
    c := byte[p_str][idx]  
    if ((c => 33) and (c =< 126))                               ' keyboard char
      c := 33 + ((c + 14) // 94)
      byte[p_str][idx] := c
    idx += 1    


pub exor(p_str, key) | len, idx, c   

'' Encrypt/decrypt buffer by xor-ing str with key
'' -- p_str is pointer to string to encode
''    * use -1 for buffer assigned to object

  if (p_str < 0)
    p_str := p_buffer                                           ' use assigned buffer

  idx := 0
    
  repeat strsize(p_str)
    c := byte[p_str][idx]
    if ((c => "A") and (c =< "Z"))
      c ^= key  
    elseif ((c => "a") and (c =< "z"))
      c ^= key 
    byte[p_str][idx++] := c 
   
  
dat { license }

{{

  Copyright (C) 2015-2016 Jon McPhalen 

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}